import click
from mpm.model import *
import pathlib
from mpm.config import ServerParamType
from columnar import columnar

@click.group(help='Add, remove, and synchronize Servers')
def server():
    pass

@server.command(help='Add a new server')
@click.pass_context
@click.argument('name')
@click.argument('path', type=click.Path(dir_okay=True, exists=True))
def add(ctx, name, path):
    ctx.obj['config'].add_server(name, path)
    ctx.obj['config'].save()
    click.echo("Added server {} in {}".format(name, path))

@server.command('list', help='List configured servers')
@click.pass_context
@click.argument('servers', type=ServerParamType(), nargs=-1)
def list_(ctx, servers):
    if len(servers) == 0:
        servers = ctx.obj['config'].servers()
    for server in servers:
        click.echo('{} ({}):'.format(server.name, server.path))
        serverPath = pathlib.Path(server.path)
        if not serverPath.exists():
            click.echo(click.style("\tServer path does not exist", fg='red'))
            continue
        pluginPath = pathlib.Path(server.path+'/plugins')
        if not pluginPath.exists():
            click.echo(click.style("\tServer plugin directory does not exist", fg='red'))
            continue
        outdatedLinks = []
        missing = []
        installed = []
        unmanaged = []
        conflicts = []
        for state in sorted(server.pluginStates(ctx.obj['config'].repositories())):
            if isinstance(state, OutdatedSymlink):
                outdatedLinks.append(state)
            elif isinstance(state, Installed):
                installed.append(state)
            elif isinstance(state, MissingVersions):
                missing.append(state)
            elif isinstance(state, UnmanagedFile):
                unmanaged.append(state)
            elif isinstance(state, SymlinkConflict):
                conflicts.append(state)

        rows = []
        for state in sorted(installed):
            rows.append([
                state.plugin.name,
                str(state.plugin.versionSpec),
                str(state.currentVersion),
                str(state.currentVersion),
            ])
        for state in sorted(outdatedLinks):
            rows.append([
                state.plugin.name,
                click.style(str(state.wantedVersion), fg='green'),
                click.style(str(state.currentVersion), fg='yellow'),
                click.style(str(state.plugin.versionSpec), fg='green')
            ])
        for state in sorted(missing):
            rows.append([
                click.style(state.plugin.name, fg='red'),
                click.style(str(state.plugin.versionSpec), fg='yellow'),
                click.style('Missing', fg='red'),
                click.style('Missing', fg='red')
            ])
        for state in sorted(unmanaged):
            rows.append([
                click.style(state.filename, fg='yellow'),
                '?',
                click.style('Unmanaged', fg='yellow', bold=True),
                ''
            ])
        for state in sorted(conflicts):
            rows.append([
                state.plugin.name,
                click.style(str(state.plugin.versionSpec), fg='yellow'),
                click.style('Symlink Conflict', fg='red', bold=True),
                click.style(str(state.plugin.versionSpec), fg='yellow'),
            ])
            click.echo("\t{}.jar".format(state.plugin.name))

        if len(rows) > 0:
            click.echo(columnar(rows, ['Plugin', 'Wanted', 'Installed',
                                       'Available'], no_borders=True))
        else:
            click.echo(click.style('No plugins configured!', fg='yellow'))

@server.command(help='Add a plugin to a server')
@click.pass_context
@click.argument('server', type=ServerParamType())
@click.argument('plugins', nargs=-1)
def add_plugin(ctx, server, plugins):
    pluginQueue = []
    for pluginSpec in plugins:
        if os.path.exists(pluginSpec):
            plugin = Plugin(pluginSpec)
            pluginSpec = PluginSpec(plugin.name, str(plugin.version))
        else:
            allVersions = []
            for repo in ctx.obj['config'].repositories():
                allVersions += repo.versionsForPlugin(pluginSpec)
            pluginSpec = PluginSpec(pluginSpec, list(reversed(sorted(allVersions)))[0])

        pluginQueue.append(pluginSpec)

    click.echo("Resolved plugin list:")
    for pluginSpec in pluginQueue:
        click.echo("\t{} {}".format(click.style(pluginSpec.name, fg='green'),
                                    click.style(pluginSpec.versionSpec,
                                    fg='yellow')))
    click.echo("Add these plugins to server {}? [y/N] ".format(server.name),
               nl=False)
    answer = click.getchar().lower()
    if answer == "y":
        for pluginSpec in pluginQueue:
            try:
                server.add_plugin(pluginSpec)
            except KeyError:
                click.echo("{} already added!".format(pluginSpec))
        ctx.obj['config'].update_server(server.name, server.config)
        ctx.obj['config'].save()
        click.echo("Added!")
    else:
        click.echo("Cancelled.")

@server.command(help="Synchronize a server's plugins")
@click.pass_context
@click.argument('servers', type=ServerParamType(), nargs=-1)
def sync(ctx, servers):
    allServers = ctx.obj['config'].servers()
    if len(servers) > 0:
        allServers = servers
    for server in allServers:
        click.echo('{} ({}):'.format(server.name, server.path))
        outdatedLinks = []
        available = []
        missing = []
        for state in sorted(server.pluginStates(ctx.obj['config'].repositories())):
            if isinstance(state, OutdatedSymlink):
                outdatedLinks.append(state)
            elif isinstance(state, Available):
                available.append(state)
            elif isinstance(state, MissingVersions):
                missing.append(state)

        rows = []

        for state in sorted(outdatedLinks):
            rows.append([
                state.plugin.name,
                click.style(str(state.wantedVersion), fg='green'),
                click.style(str(state.currentVersion), fg='yellow'),
                click.style(str(state.plugin.versionSpec), fg='green')
            ])

        for state in sorted(available):
            rows.append([
                state.plugin.name,
                click.style(str(state.wantedVersion), fg='green'),
                click.style('Missing', fg='red'),
                click.style(str(state.plugin.version), fg='green')
            ])

        click.echo("Missing plugins:")
        for state in sorted(missing):
            rows.append([
                click.style(state.plugin.name, fg='red'),
                click.style(str(state.plugin.versionSpec), fg='yellow'),
                click.style('Missing', fg='red'),
                click.style('Missing', fg='red')
            ])

        if len(rows) > 0:
            click.echo(columnar(rows, ['Plugin', 'Wanted', 'Installed',
                                       'Available'], no_borders=True))

        if len(outdatedLinks) > 0 or len(available) > 0:
            click.echo("Apply changes? [y/N]", nl=False)
            answer = click.getchar().lower()
            if answer == "y":
                with click.progressbar(length=len(available)+len(outdatedLinks),
                                       label='Synchronizing') as bar:
                    i = 0
                    for state in available:
                        server.installVersion(state.plugin)
                        server.updateSymlinkForPlugin(state.plugin, state.plugin.version)
                        i += 1
                        bar.update(i)
                    for state in outdatedLinks:
                        server.updateSymlinkForPlugin(state.plugin, state.wantedVersion)
                        i += 1
                        bar.update(i)
            else:
                click.echo("Not applying changes.")
        else:
            click.echo("No changes to apply.")
